/**********************************************************
*
*   PolyCubicSpline.java
*
*   Class for performing an interpolation on the tabulated
*   function y = f(x1,x2, x3 .... xn) using a natural cubic splines
*   Assumes second derivatives at end points = 0 (natural spines)
*
*   WRITTEN BY: Dr Michael Thomas Flanagan
*
*   DATE:	    15 February 2006,
*   UPDATES:    9 June 2007, 27 July 2007
*
*   DOCUMENTATION:
*   See Michael Thomas Flanagan's Java library on-line web page:
*   http://www.ee.ucl.ac.uk/~mflanaga/java/PolyCubicSpline.html
*   http://www.ee.ucl.ac.uk/~mflanaga/java/
*
*   Copyright (c) February 2006, July 2007  Michael Thomas Flanagan
*
*   PERMISSION TO COPY:
*   Permission to use, copy and modify this software and its documentation for
*   NON-COMMERCIAL purposes is granted, without fee, provided that an acknowledgement
*   to the author, Michael Thomas Flanagan at www.ee.ucl.ac.uk/~mflanaga, appears in all copies.
*
*   Dr Michael Thomas Flanagan makes no representations about the suitability
*   or fitness of the software for any or for a particular purpose.
*   Michael Thomas Flanagan shall not be liable for any damages suffered
*   as a result of using, modifying or distributing this software or its derivatives.
*
***************************************************************************************/

package flanagan.interpolation;

import flanagan.math.Fmath;
import java.lang.reflect.Array;

public class PolyCubicSpline{

    	private int nDimensions = 0;   	    // number of the dimensions of the tabulated points array, y=f(x1,x2,x3 . . xn), i.e. n
    	private Object fOfX = null;         // tabulated values of y = f(x1,x2,x3 . . fn)
    	                                    // as a multidimensional array of double [x1 length][x2 length] ... [xn length]
    	private Object internalArray = null;// Object to store second derivatives
    	                                    // as a multidimensional array of Objects the innermost three layers being double arrays
    	private Object xArrays = null;      // The variable arrays x1, x2, x3 . . . xn
    	                                    // packed as a multidemensional array of double [][]
    	                                    // where xArrays[0] = array of x1 values, xArrays[1] = array of x2 values etc
        private double yValue = 0.0D;       // returned interpolated value
    	private boolean calculationDone = false;    // = true when derivatives calculated


    	// Constructor
    	public PolyCubicSpline(Object xArrays, Object fOfX){

    	    this.fOfX = Fmath.copyObject(fOfX);
    	    this.xArrays = Fmath.copyObject(xArrays);

    	    // Calculate fOfX array dimension number
    	    Object internalArray = Fmath.copyObject(fOfX);
    	    this.nDimensions = 1;
            while(!((internalArray  =  Array.get(internalArray, 0)) instanceof Double))nDimensions++;

            // Repack xArrays as 2 dimensional array if entered a single dimensioned array for a simple cubic spline
            if(this.xArrays instanceof double[] && this.nDimensions == 1){
                double[][] xArraysTemp = new double[1][];
                xArraysTemp[0] = (double[])this.xArrays;
                this.xArrays = (Object)xArraysTemp;
            }
            else{
               if(!(this.xArrays instanceof double[][]))throw new IllegalArgumentException("xArrays should be a two dimensional array of doubles");
            }
        }

    	//  Interpolation method
    	public double interpolate(double[] unknownCoord){

    	    int nUnknown = unknownCoord.length;
    	    if(nUnknown!=this.nDimensions)throw new IllegalArgumentException("Number of unknown value coordinates, " + nUnknown + ", does not equal the number of tabulated data dimensions, " + this.nDimensions);

            int kk = 0;
            double[][] xArray = (double[][])this.xArrays;
            switch(this.nDimensions){
                case 0: throw new IllegalArgumentException("data array must have at least one dimension");
                case 1: // If fOfX is one dimensional perform simple cubic spline
                        CubicSpline cs = new CubicSpline(xArray[0], (double[])this.fOfX);
                        if(this.calculationDone)cs.setDeriv((double[])this.internalArray);
                        this.yValue = cs.interpolate(unknownCoord[0]);
                        if(!this.calculationDone){
                            double[] deriv = cs.getDeriv();
                            this.internalArray = (Object)deriv;
                            this.calculationDone = true;
                        }
                        break;
                case 2: // If fOfX is two dimensional perform bicubic spline
                        BiCubicSpline bcs = new BiCubicSpline(xArray[0], xArray[1], (double[][])this.fOfX);
                        if(this.calculationDone)bcs.setDeriv((double[][])this.internalArray);
                        this.yValue = bcs.interpolate(unknownCoord[0], unknownCoord[1]);
                        if(!this.calculationDone){
                            double[][] deriv = bcs.getDeriv();
                            this.internalArray = (Object)deriv;
                            this.calculationDone = true;
                        }
                        break;
                case 3: // If fOfX is three dimensional perform tricubic spline
                        TriCubicSpline tcs = new TriCubicSpline(xArray[0], xArray[1], xArray[2], (double[][][])this.fOfX);
                        if(this.calculationDone)tcs.setDeriv((double[][][])this.internalArray);
                        this.yValue = tcs.interpolate(unknownCoord[0], unknownCoord[1], unknownCoord[2]);
                        if(!this.calculationDone){
                            double[][][] deriv = tcs.getDeriv();
                            this.internalArray = (Object)deriv;
                            this.calculationDone = true;
                        }
                        break;
                default:// If fOfX is greater than three dimensional, recursively call PolyCubicSpline
                        //  with, as arguments, the n1 fOfX sub-arrays, each of (number of dimensions - 1) dimensions,
                        //  where n1 is the number of x1 variables.
                        Object obj = fOfX;
                        int dimOne = Array.getLength(obj);
                        double[] csArray = new double [dimOne];
                        double[][] newXarrays= new double[this.nDimensions-1][];
                        double[] newCoord = new double[this.nDimensions-1];
                        for(int i=0; i<this.nDimensions-1; i++){
                            newXarrays[i] = xArray[i+1];
                            newCoord[i] = unknownCoord[i+1];
                        }
                        Object[] objDeriv = new Object[dimOne];
                        if(calculationDone)objDeriv = (Object[])this.internalArray;
                        for(int i=0; i<dimOne; i++){
                            Object objT = (Object)Array.get(obj, i);
                            PolyCubicSpline pcs = new PolyCubicSpline(newXarrays, objT);
                            if(this.calculationDone)pcs.setDeriv(objDeriv[i]);
                            csArray[i] = pcs.interpolate(newCoord);
                            if(!this.calculationDone)objDeriv[i]  = pcs.getDeriv();
                        }
                        this.internalArray = (Object)objDeriv;
                        this.calculationDone = true;


                        // Perform simple cubic spline on the array of above returned interpolates
                        CubicSpline ncs = new CubicSpline(xArray[0], csArray);
            	    	this.yValue = ncs.interpolate(unknownCoord[0]);
            }

            return this.yValue;
    	}

    	// Set derivatives (internal array)
    	public void setDeriv(Object internalArray){
    	    this.internalArray = internalArray;
    	}

    	// Get derivatives (internal array)
    	public Object getDeriv(){
    	    return this.internalArray;
    	}


}

